/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.databinding.jaxb;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.wsdl.Binding;
import javax.wsdl.BindingOperation;
import javax.wsdl.Definition;
import javax.wsdl.Fault;
import javax.wsdl.Input;
import javax.wsdl.Output;
import javax.wsdl.Port;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.schema.Schema;
import javax.wsdl.extensions.schema.SchemaImport;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;

import org.ow2.petals.databinding.jaxb.serialization.ServiceMarshaller;
import org.ow2.petals.databinding.jaxb.serialization.ServiceUnmarshaller;
import org.ow2.petals.databinding.jaxb.service.Message;
import org.ow2.petals.databinding.jaxb.service.Operation;
import org.ow2.petals.databinding.jaxb.service.Service;
import org.ow2.petals.databinding.jaxb.tools.Util;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.ibm.wsdl.PartImpl;
import com.ibm.wsdl.extensions.schema.SchemaImportImpl;
import com.ibm.wsdl.extensions.soap.SOAPOperationImpl;
import com.ibm.wsdl.extensions.soap12.SOAP12OperationImpl;

/**
 * This class is a Factory to create an instance of the Service class. 
 * <p>
 * The creation of the Service object is based on WSDL file and the generated 
 * JAXB classes.
 * </p>
 * 
 * @author Mohammed EL JAI - EBM WebSourcing
 */
public class ServiceFactory {


    /** The single object **/
    private static ServiceFactory wsdl2JbiFactory;

    /**
     * Static method to get a single instance of this class
     * @return An instance of this class
     */
    public static ServiceFactory getInstance() {
        if(wsdl2JbiFactory == null) 
            return new ServiceFactory();
        return wsdl2JbiFactory;
    }


    /**
     * Method to create the service object
     * 
     * @param wsdlUrl
     *            : The WSDL URL
     * @param jaxbClassJar
     *            : The JAR URL that gathers the generated JAXB classes
     * @param serviceClassLoader
     *            : The service class loader specified
     * @param serviceClass
     * @return An instance of the Service class
     * @throws DataBindingException
     */
    public Service crateService(
            String wsdlUrl, String jaxbClassJar, 
            ClassLoader serviceClassLoader, Class<?> serviceClass) 
    throws DataBindingException {

        Service _service = new Service();
        setWsdlServiceInfo(wsdlUrl, _service);
        setJaxbInfo(jaxbClassJar, _service, serviceClassLoader);
        addMethodSignature(_service, serviceClass);

        return _service;
    }


    /**
     * Set WSDL informations to the Service object given in parameter.
     * @param wsdlUrl : The WSDL URL
     * @param _service : The service object
     * @throws DataBindingException 
     */
    @SuppressWarnings("unchecked")
    private void setWsdlServiceInfo(String wsdlUrl, Service _service) 
    throws DataBindingException {

        WSDLFactory wsdlFactory = null;
        Definition definition = null;

        try {
            wsdlFactory = WSDLFactory.newInstance();
        } catch (WSDLException e1) {
            throw new DataBindingException(
                    "Error occurred during the creation of the WSDL factory.", e1);
        }

        // Browse the WSDL Elements
        WSDLReader wsdlReader = wsdlFactory.newWSDLReader();
        try {
            definition = wsdlReader.readWSDL(wsdlUrl);
            Map<QName,javax.wsdl.Service> services = definition.getServices();
            Set<QName> serviceNames = services.keySet();

            // VZ: PETALSBCEJB-7: WSDL are not handled correctly if they contain
            // schema imports
            // Get all the schemas
            List<?> exts = definition.getTypes().getExtensibilityElements();
            List<Schema> schemas = new ArrayList<Schema>();
            for (Object o : exts) {
                if (o instanceof Schema) {
                    Schema schema = (Schema) o;
                    schemas.add(schema);

                    for (Object coll : schema.getImports().values()) {
                        if (coll instanceof Collection) {
                            for (Object oo : (Collection<?>) coll) {
                                if (oo instanceof SchemaImport)
                                    schemas.add(((SchemaImport) oo).getReferencedSchema());
                            }
                        }
                    }
                }
            }

            // Register the Classes with their package in a map
            // Process all the schemas for that
            for (Schema schema : schemas) {
                String tns = schema.getElement().getAttribute("targetNamespace");
                NodeList children = schema.getElement().getChildNodes();

                for (int i = 0; i < children.getLength(); i++) {
                    Node child = children.item(i);
                    if (child instanceof Element && "complexType".equals(child.getLocalName())) {

                        String name = ((Element) child).getAttribute("name");
                        _service.addClassMapping(name,
                                com.sun.tools.xjc.api.XJC.getDefaultPackageName(tns));
                    }
                }
            }
            // End of fix
            // VZ: PETALSBCEJB-7: WSDL are not handled correctly if they contain
            // schema imports

            javax.wsdl.Service service = services.get(serviceNames.iterator().next());
            _service.setServiceName(service.getQName());

            Map<String, Port> ports = service.getPorts();
            Set<String> portNames = ports.keySet();

            for (String portName : portNames) {

                // If the service has more than one port, choose the soap port
                if (ports.size() > 1 && !portName.endsWith("Soap")) 
                    continue;

                Port port = ports.get(portName);
                Binding binding = port.getBinding();
                _service.setInterfaceName(binding.getPortType().getQName());
                List<BindingOperation> bindingOperations = binding.getBindingOperations();

                // Browse the binding operations
                for (BindingOperation bindingOperation : bindingOperations) {

                    // Create service operation
                    Operation serviceOperation = new Operation();
                    serviceOperation.setServiceOfOperation(_service);
                    javax.wsdl.Operation operation = bindingOperation.getOperation();
                    serviceOperation.setOperationName(operation.getName());

                    // Create input message 
                    Message inputMessage = new Message();
                    Input input = operation.getInput();
                    javax.wsdl.Message message = input.getMessage();

                    if (message.getParts().size() != 1)
                        throw new DataBindingException("The message element has more than one part, " +
                                "only the document/literal wrapped " +
                        "WSDL style is supported ");

                    PartImpl partImpl = (PartImpl) message.getParts().values().iterator().next();
                    String partName = partImpl.getName();

                    inputMessage.setMessageName(message.getPart(partName).getElementName());
                    inputMessage.setMessageTypeName(getMessageType(definition, message.getPart(partName).getElementName()));
                    serviceOperation.setInputMessage(inputMessage);

                    // Create output message
                    Message outputMessage = new Message();
                    Output output = operation.getOutput();
                    if (output != null) {
                        message = output.getMessage();

                        if (message.getParts().size() != 1)
                            throw new DataBindingException("The message element has more than one part, " +
                                    "only the document/literal wrapped " +
                            "WSDL style is supported ");

                        partImpl = (PartImpl) message.getParts().values().iterator().next();
                        partName = partImpl.getName();

                        outputMessage.setMessageName(message.getPart(partName).getElementName());
                        outputMessage.setMessageTypeName(getMessageType(definition, message.getPart(partName).getElementName()));
                        serviceOperation.setOutputMessage(outputMessage);
                    }

                    // Create the faults messages
                    Map<?,?> faults = operation.getFaults();
                    for (String faultName : (Set<String>) faults.keySet()) {
                        Fault fault = (Fault) faults.get(faultName);
                        Message faultMessage = new Message();
                        message = fault.getMessage();
                        partImpl = (PartImpl) message.getParts().values().iterator().next();
                        partName = partImpl.getName();
                        faultMessage.setMessageName(message.getQName());
                        faultMessage.setMessageTypeName(getMessageType(definition, message.getPart(partName).getElementName()));
                        serviceOperation.getOperationFaults().add(faultMessage);
                    }

                    // Get the soap action attribute from the extensible elements
                    List<?> extensions = bindingOperation.getExtensibilityElements();
                    for (Object extension : extensions) {
                        if (extension instanceof SOAPOperationImpl) {
                            SOAPOperationImpl soapOperation = (SOAPOperationImpl) extension;
                            String soapAction = soapOperation.getSoapActionURI();
                            serviceOperation.setSoapActionName(soapAction);
                        }
                        else if (extensions instanceof SOAP12OperationImpl) {
                            SOAP12OperationImpl soapOperation = (SOAP12OperationImpl) extension;
                            String soapAction = soapOperation.getSoapActionURI();
                            serviceOperation.setSoapActionName(soapAction);
                        }
                    }

                    // Add the operation to the service
                    _service.getOperations().add(serviceOperation);
                }
            }
        }

        catch (Exception e) {
            throw new DataBindingException("WSDL parsing error, can't get service inofrmations" +
                    "From the WSDL", e);
        }

    }


    /**
     * Set JAXB informations to the Service object given in parameter. 
     * @param jaxbClassJar : The JAR URL that gathers the generated JAXB classes 
     * @param _service : The service object
     * @param serviceClassLoader : The service class loader 
     * @param encodingType : parameter to configure marshalling/unmarshalling encoding. It can be not
     * specified.
     * @throws DataBindingException 
     */
    private void setJaxbInfo(String jaxbClassJar, Service _service, ClassLoader serviceClassLoader) 
    throws DataBindingException {

        ServiceMarshaller serviceMarshaller = new ServiceMarshaller();
        serviceMarshaller.setClassPackageMapping(_service.getClassPackageMapping());

        ServiceUnmarshaller serviceUnmarshaller = new ServiceUnmarshaller();

        try {

            ClassLoader jaxbClassLoader = 
                new URLClassLoader(new URL[] {new URL(jaxbClassJar)}, this.getClass().getClassLoader());
            serviceMarshaller.setJaxbClassLoader(jaxbClassLoader);
            serviceUnmarshaller.setServiceClassLoader(serviceClassLoader);
            serviceUnmarshaller.setJaxbClassLoader(jaxbClassLoader);

            // Registry to store marshalling objects (Marshallern UnMarshaller..), it's used  
            // to avoid the creation of multiples instances of the same objects
            Registry registry = new Registry();

            List<Operation> serviceOperations = _service.getOperations();
            // Browse the service operations
            for (Operation serviceOperation : serviceOperations) {

                // Add JAXB classes correspond to the input message
                Message inputMessage = serviceOperation.getInputMessage();
                String messageTypeName = Util.upperFirstCase(
                        inputMessage.getMessageTypeName().getLocalPart());

                String packageName = 
                    com.sun.tools.xjc.api.XJC.getDefaultPackageName(inputMessage.getMessageTypeName().getNamespaceURI());
                String jaxbMessageClassName = packageName + "." + messageTypeName;

                // Add marshalling/unMarshalling objects at the input message
                if (! registry.getMarshallingObjects().
                        containsKey(inputMessage.getMessageTypeName().getNamespaceURI())) 
                    registry.addEntry(inputMessage.getMessageTypeName().getNamespaceURI(),
                            packageName, jaxbClassLoader);

                // Construct the java representation of the input message 
                Class<?> jaxbMessageClass = jaxbClassLoader.loadClass(jaxbMessageClassName);
                inputMessage.setMessageJaxbClass(jaxbMessageClass);
                Field[] fields = jaxbMessageClass.getDeclaredFields();
                for (Field field : fields) 
                    inputMessage.getMessageParts().put(field.getName(), field.getType());

                // Construct the java representation of the output message
                Message outputMessage = serviceOperation.getOutputMessage();
                if (outputMessage != null) {
                    messageTypeName = Util.upperFirstCase(
                            outputMessage.getMessageTypeName().getLocalPart());
                    packageName = 
                        com.sun.tools.xjc.api.XJC.getDefaultPackageName(outputMessage.getMessageTypeName().getNamespaceURI());
                    jaxbMessageClassName = packageName + "." + messageTypeName;

                    // Add marshalling/unMarshalling objects at the output message
                    if (! registry.getMarshallingObjects().
                            containsKey(outputMessage.getMessageTypeName().getNamespaceURI())) 
                        registry.addEntry(outputMessage.getMessageTypeName().getNamespaceURI(),
                                packageName, jaxbClassLoader);

                    // Construct the java representation of the output message 
                    jaxbMessageClass = jaxbClassLoader.loadClass(jaxbMessageClassName);
                    outputMessage.setMessageJaxbClass(jaxbMessageClass);
                    fields = jaxbMessageClass.getDeclaredFields();
                    for (Field field : fields) {
                        String filedName = field.getName();
                        if (filedName.equals("_return")) 
                            filedName = filedName.substring(1);
                        outputMessage.getMessageParts().put(filedName, field.getType());
                    }
                }

                // Construct the JAVA representation for the fault message
                List<Message> faults = serviceOperation.getOperationFaults();
                for (Message faultMessage : faults) {
                    messageTypeName = Util.upperFirstCase(
                            faultMessage.getMessageTypeName().getLocalPart());
                    packageName = 
                        com.sun.tools.xjc.api.XJC.getDefaultPackageName(faultMessage.getMessageTypeName().getNamespaceURI());
                    jaxbMessageClassName = packageName + "." + messageTypeName;

                    // Add marshalling/unMarshalling objects at the input message
                    if (! registry.getMarshallingObjects().
                            containsKey(faultMessage.getMessageTypeName().getNamespaceURI())) 
                        registry.addEntry(faultMessage.getMessageTypeName().getNamespaceURI(),
                                packageName, jaxbClassLoader);

                    // Construct the java representation of the output message 
                    jaxbMessageClass = jaxbClassLoader.loadClass(jaxbMessageClassName);
                    faultMessage.setMessageJaxbClass(jaxbMessageClass);
                    fields = jaxbMessageClass.getDeclaredFields();
                    for (Field field : fields) 
                        faultMessage.getMessageParts().put(field.getName(), field.getType());

                }

            }
            _service.setRegistry(registry);
            _service.setServiceUnmarshaller(serviceUnmarshaller);
            _service.setServiceMarshaller(serviceMarshaller);

        } catch (Exception e) {
            throw new DataBindingException(" Can't extract the DataBinding inofrmations", e);
        } 

    }



    /**
     * Set method signatures to the service object given in parameter
     * @param service : The service object.
     * @param serviceClass : The JAVA interface class of the service
     */
    private void addMethodSignature(Service service, Class<?> serviceClass) {

        List<Operation> serviceOperations = service.getOperations();
        for (Operation serviceOperation : serviceOperations) {
            Method[] methods = serviceClass.getMethods();
            for (Method method : methods) {
                if (method.getName().equalsIgnoreCase(serviceOperation.getOperationName())) {
                    // Set the parameters Type
                    serviceOperation.setParameterTypes(method.getParameterTypes());
                    serviceOperation.setGenericParameterTypes(method.getGenericParameterTypes());
                    // Set the return type
                    if (! method.getReturnType().equals(void.class)) {
                        serviceOperation.setReturnType(method.getReturnType());
                        serviceOperation.setGenericReturnType(method.getGenericReturnType());
                    }
                    // Set the Exceptions type
                    serviceOperation.setExceptionType(method.getExceptionTypes());
                }
            }
        }
    }


    /**
     * Search message type QName of the WSDL message from the WSDL definition
     * @param wsdlDefintion : The root WSDL Element
     * @param messageElement : The message element name
     * @return : The message type name of the message name given in parameter
     */
    private static QName getMessageType(Definition wsdlDefintion, QName messageElement) {

        List<?> exts = wsdlDefintion.getTypes().getExtensibilityElements();
        for (Object o : exts){
            Schema rootSchema= (Schema) o;
            QName messageType =  getMessageTypeFromSchema(rootSchema, messageElement, wsdlDefintion);
            if (messageType != null)
                return messageType;
        }
        return null;   


    }

    /**
     * Get the type name of the message element name 
     * @param schema : The schema that can be contains the message type
     * @param messageElement : The message element name
     * @param Definition wsdlDefintion : The root element of the WSDL
     * @return The massage type name
     */
    @SuppressWarnings("unchecked")
    private static QName getMessageTypeFromSchema(Schema schema, QName messageElement, Definition wsdlDefintion) {

        NodeList nodeList = schema.getElement().getChildNodes();
        for (int i =0; i < nodeList.getLength(); i++){
            Node currentNode = nodeList.item(i);
            if ("element".equals(currentNode.getLocalName())){
                NamedNodeMap atts2 = currentNode.getAttributes();
                Node name = atts2.getNamedItem("name");
                String nodeValue = name.getNodeValue();
                String nodeValueNameSpace;
                String nodeValueLocalPart;
                if (!nodeValue.contains(":")) {
                    nodeValueNameSpace = getNameSpaceForPrefix(schema,"targetNamespace", wsdlDefintion);
                    nodeValueLocalPart = nodeValue;
                }
                else {
                    String prefix = nodeValue.substring(0, nodeValue.indexOf(':'));
                    nodeValueNameSpace = getNameSpaceForPrefix(schema,prefix, wsdlDefintion);
                    nodeValueLocalPart = nodeValue.substring(nodeValue.indexOf(':'), nodeValue.length());
                }
                QName currentElementName = new QName(nodeValueNameSpace,nodeValueLocalPart);
                if (messageElement.equals(currentElementName)) {
                    Node type = atts2.getNamedItem("type");
                    if (type == null)
                        return currentElementName;
                    else {
                        String typeValue = type.getNodeValue();
                        String typeValueNameSpace;
                        String typeValueLocalPart;
                        if (!typeValue.contains(":")) {
                            typeValueNameSpace = getNameSpaceForPrefix(schema,"targetNamespace", wsdlDefintion);
                            typeValueLocalPart = typeValue;
                        }
                        else {
                            String prefix = typeValue.substring(0, typeValue.indexOf(':'));
                            typeValueNameSpace = getNameSpaceForPrefix(schema,prefix, wsdlDefintion);
                            typeValueLocalPart = typeValue.substring(typeValue.indexOf(':') + 1, typeValue.length());
                        }
                        return new QName(typeValueNameSpace, typeValueLocalPart);
                    }

                }

            }
        }

        // Search the message type name in the imported schemas
        Map<String, Vector<SchemaImportImpl>> imports = schema.getImports();
        Set<String> keys = imports.keySet();
        QName messageTypeName;
        for (String key : keys) {
            Vector<SchemaImportImpl> importedSchemas = imports.get(key);
            for (SchemaImportImpl importedSchema : importedSchemas) {
                if (importedSchema.getReferencedSchema() != null) {
                    messageTypeName = getMessageTypeFromSchema(importedSchema.getReferencedSchema(), 
                            messageElement, wsdlDefintion);
                    if (messageTypeName != null)
                        return messageTypeName;
                }
            }
        }

        return null;
    }



    /**
     * Get the name space associated to the prefix from the schema attributes elements
     * @param schema : Schema that contains the name space
     * @param prefix : Prefix of the searched name space
     * @return the name space associated to the prefix given in parameter
     */
    private static String getNameSpaceForPrefix(Schema schema, String prefix, Definition wsdlDefinition) {

        NamedNodeMap atts = schema.getElement().getAttributes();
        Node targetNS;
        if (prefix.equals("targetNamespace"))
            targetNS = atts.getNamedItem("targetNamespace");
        else
            targetNS = atts.getNamedItem("xmlns:"+prefix);

        if (targetNS == null) {


            if (prefix.equals("targetNamespace"))
                return wsdlDefinition.getTargetNamespace();
            else
                return (String) wsdlDefinition.getNamespaces().get(prefix);
        }

        return targetNS.getNodeValue();


    }


}
